const plugin = function (babel: {
  types: any
}) {
  const t = babel.types

  // 这些变量需要在每个 program 里重置
  const invokedApis: Map<string, string> = new Map()
  let vnmfName: string
  let needDefault: boolean

  let referVnmf: any[]

  return {
    name: 'babel-plugin-transform-vnmf-api',
    visitor: {
      ImportDeclaration (ast, state) {
        const packageName = state.opts.packageName
        const apis = state.opts.apis
        if (ast.node.source.value !== packageName) return

        ast.node.specifiers.forEach(node => {
          if (t.isImportDefaultSpecifier(node)) {
            needDefault = true
            vnmfName = node.local.name
          } else if (t.isImportSpecifier(node)) {
            const propertyName = node.imported.name
            if (apis.has(propertyName)) { // 记录api名字
              ast.scope.rename(node.local.name)
              invokedApis.set(propertyName, node.local.name)
            } else { // 如果是未实现的api 改成Vnmf.xxx
              needDefault = true
              const localName = node.local.name
              const binding = ast.scope.getBinding(localName)
              const idn = t.identifier(vnmfName)
              referVnmf.push(idn)
              binding && binding.referencePaths.forEach(reference => {
                reference.replaceWith(
                  t.memberExpression(
                    idn,
                    t.identifier(propertyName)
                  )
                )
              })
            }
          }
        })
      },
      MemberExpression (ast, state) {
        /* 处理Vnmf.xxx */
        const apis = state.opts.apis
        const isVnmf = t.isIdentifier(ast.node.object, { name: vnmfName })
        const property = ast.node.property
        let propertyName: string | null = null
        let propName = 'name'

        if (!isVnmf) return

        // 兼容一下 Vnmf['xxx']
        if (t.isStringLiteral(property)) {
          propName = 'value'
        }
        propertyName = property[propName]

        if (!propertyName) return

        // 同一api使用多次, 读取变量名
        if (apis.has(propertyName)) {
          const parentNode = ast.parent
          const isAssignment = t.isAssignmentExpression(parentNode) && parentNode.left === ast.node

          if (!isAssignment) {
            let identifier: any
            if (invokedApis.has(propertyName)) {
              identifier = t.identifier(invokedApis.get(propertyName)!)
            } else {
              const newPropertyName = ast.scope.generateUid(propertyName)
              invokedApis.set(propertyName, newPropertyName)
              /* 未绑定作用域 */
              identifier = t.identifier(newPropertyName)
            }
            ast.replaceWith(identifier)
          }
        } else {
          needDefault = true
        }
      },
      Program: {
        enter (ast) {
          needDefault = false
          referVnmf = []
          invokedApis.clear()

          vnmfName = ast.scope.getBinding('Vnmf')
            ? ast.scope.generateUid('Vnmf')
            : 'Vnmf'
        },
        exit (ast, state) {
          // 防止重复引入
          let isVnmfApiImported = false
          referVnmf.forEach(node => {
            node.name = vnmfName
          })

          ast.traverse({
            ImportDeclaration (ast) {
              const packageName = state.opts.packageName
              const isImportingVnmfApi = ast.node.source.value === packageName
              if (!isImportingVnmfApi) return
              if (isVnmfApiImported) return ast.remove()
              isVnmfApiImported = true
              const namedImports = Array.from(invokedApis.entries()).map(([imported, local]) => t.importSpecifier(t.identifier(local), t.identifier(imported)))
              if (needDefault) {
                const defaultImport = t.importDefaultSpecifier(t.identifier(vnmfName))
                ast.node.specifiers = [
                  defaultImport,
                  ...namedImports
                ]
                needDefault = false
              } else {
                ast.node.specifiers = namedImports
              }
            }
          })
        }
      }
    }
  }
}
export default plugin
